/*

DESCRIPTION:    control firmware for digital barometric pressure altimeter 
                using TE MS5607.

                BALT, short for Barometric ALTimeter, is intended to be a simple
                hand-held unit to display the altitude (elevation) of the user
                using ambient atmospheric pressure. It is powered with a single
                CR2032 3V lithium coin cell. The PIC communicates with the
                MS5607 and LCD modules using Serial Peripheral Interface (SPI)
                configured for 400kHz.

                Features:
                + 8 digit, 12mm (0.472") tall, high contrast transflective LCD
                + User activated back-light for nighttime use
                + Displays altitude, accumulated climb/descent, vertical speed,
                  temperature, pressure, and compensation
                + English or metric units
                + Compensation settable in feet, meters, inches of mercury (Hg),
                  or millibars (mbar)
                + Sleep and deep sleep (off) to conserve battery life
                + Adjustable off time from 0 (always on) to 12 hours
                + Operating temperature range -20 to +70C

                Accuracy: linear estimate of +/- 1.5% of displayed altitude 
                based on measurements taken between MSL and 2,133M (7,000F) 
                using proximate AWOS (NWS Automated Weather Observation Stations)
                barometric settings. However, no accuracy specification has been 
                specified for this design because:
                    + Barometric altimeters are based on the International 
                      Standard Atmosphere (ISA) which uses a Mean Sea Level
                      (MSL) pressure of 1013.25mbar or 29.92" Hg and a change in
                      pressure of 1" Hg per 1,000 feet (up to 18,000F).
                    + The MS5607 barometric sensor measures pressure in mbar.
                    + The compensation and initial altitude solution are 
                      calculated in mbar/meters.
                    + There is no single standard in mbar for pressure change
                      per meter (they are based on temperature). Hence, this
                      implementation uses an equivalent value based on the 
                      English standard of 1,000F per inch of Hg (ref. constant
                      S_QY_ALT_MBAR_METR).
                    + The conversion algorithms are limited to three decimal
                      places and this results in some loss of resolution and
                      discrepancy between English & metric values.
                    + The unit senses pressure & temperature based on where it
                      is positioned; there is no guarantee of proper sensing of
                      ambient atmosphere.

                WARRANTY & DISCLAIMER: THIS UNIT IS NOT WARRANTED FOR ANY USE
                AND THE USER ASSUMES ALL RISK & RESPONSIBILITY IN ESTABLISHING
                ITS SUITABILITY & FITNESS FOR A PARTICULAR USE.

                Because a battery with very limited capacity is used power 
                management is crucial for reasonable battery life. The PIC's 
                sleep feature is used to idle the processor between baro sensor
                acquisitions & display updates. The constant CFG_WDT_CON sets
                the sleep interval. The power management features of this PIC 
                are also used to disable all unused peripherals to further 
                reduce energy consumption.

                During periods of extended inactivity the unit is placed into 
                deep sleep. Inactivity is defined as no button presses and the
                LCD back light off. Deep sleep is not a PIC feature; just a 
                descriptor for minimal energy consumption. In deep sleep the LCD
                and baro sensor are powered off, the charge pump is powered off,
                the PIC's watchdog disabled, and the PIC placed in sleep mode. 
                The PIC is awakened by a button press via the IOC feature.

                In deep sleep current consumption is reduced to ~ 32uA, which
                for a 200mAh battery would provide for a theoretical ~ 260 day
                battery life (with no use). Actual life greatly depends on the 
                amount of use. The constant S_IN_OFF sets the default timeout 
                for deep sleep.

COMPILE:        Compiled & linked with XC8 v2.31 and pack version 1.5.133.
                + C99 standard for compile & link.
                + Optimization level 2.
                + Build version control managed with build.sh that maintains 
                + build.h.

KNOWN BUGS:     + The PIC's system oscillator tolerance is +/- 5% at temperature
                  extremes and tends to speed up at low temperatures. This
                  effects button long press detection and can make it difficult 
                  for the user to enter setup or turn on the back light.
                + Vertical speed display can be erratic, particularly when the
                  unit is motionless or navigating with the buttons. Filtering
                  work is required.
                + The display behavior for climb/descent/vertical speed at the
                  value limits of the display have not been validated.

COPYRIGHT:      (c) 2021, Brian Cornell, all rights reserved.

DEPENDENCIES:   see include file listing.

 TARGET MCU:    PIC16LF18326

HW REVISION:    n/a

HW RELEASE:     210120

VERSION:        2.0

MODIFICATION HISTORY:

    17 Jan 21, BWC, version 2.0. Changed MCU to 'L' low power version.
        + Pin-out updated for HW release 210120: re-mapped DOWN pin RA4->RA3
          (RA3 is Vpp & input only). RA4 mapped as charge pump / LDO enable
          pin.
        + Added charge pump / LDO power management to firmware to power-down
          when in deep sleep. Boot & power-up now start charge pump, delay,
          and then power-up Baro & LCD to reduce battery stress and minimize
          the chances that the PIC resets when the battery is near end-life.
        + Created macros PIN_IOC_UP & PIN_IOC_DOWN to replace hard-coded IOC bit 
          flag objects.
        + Changed OSR from 256->512 to reduce display jitter.
        + Reduced repeat button time delay (S_TM_INCR_RP) from 350->200mS to
          make large value changes faster.
        + Changed definition of constant S_RT_VS_RATE from value of 15 (# 
          samples per minute) to calculation based on system constant 
          S_TM_SAMPLE so that future changes to the sampling interval correctly
          update the vertical speed calculation.

    3 Jan 21, BWC, version 1.1, functional improvements & bug fixes:
        + Clear all button presses when exiting version # display @ boot to 
          prevent display units/value/setup from engaging upon entering RUN loop
          (could occur if user presses down button before up).
        + Change MS5607 OSR (Over Sampling Rate) to 256 to reduce power
          consumption.
        + Change command constant for temperature/pressure acquisition in 
          ms5607.c from specific OSR command to S_CD_MS5607_D2 (temp) & 
          S_CD_MS5607_D2 (pres) and assign OSR command to it in ms5607.h so 
          future changes do not touch source; only include file.
        + Corrected hard-coded constants in setup single/repeat button press
          block.
        + Added user configurable deep-sleep timeout. Timeout is set in setup. A
          value of zero disables timeout. Eliminated use of bit flags to direct
          binary setup flow in favor of byte-mapped bits to allow for more than 
          two setup options in an efficient manner.
        + Modified climb/descent accumulator block. Changed constant 
          S_LM_AC_S_MIN from 1 to 2 meters as minimum change to consider (e.g. 
          10 feet since evaluation is greater than). Based on observation this 
          appears to be the variation from sample-sample when not moving. Added
          hysteresis algorithm to further reduce / eliminate spurious climb / 
          descent accumulations due to sensor.

    27 Dec 20, BWC, first production release (1.0)

    21 Dec 20, BWC, adapted for hardware release 201218.

    9 Dec 20, BWC, created project.
    

*/


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * COMPILER & CLIB INCLUDES.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <pic.h>
#include <xc.h>

/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * FIRMWARE INCLUDES:
 * + _xtal_freq.h:  sets the constant _XTAL_FREQ to the system oscillator
 *                  frequency (Fosc).  Used by the __delayxxx macros.
 * + binary.h:      Boolean constants shared across firmware projects.
 * + config.h:      PIC config registers.
 * + spi.h:         spi module include file.
 * + ms567.h:       main include file for this file.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
#include <binary.h>
#include "config.h"
#ifndef _XTAL_FREQ
#include "_xtal_freq.h"
#endif
#include "spi.h"
#include "ms5607.h"


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 * MODULE DATA.
 * + t_msc is the system mS timer used by the interrupt handler for timer6.
 *   Counts the number of mS in a second in 10mS increments.
 * + t_sc is the system seconds timer used by the interrupt handler for timer6.
 *   Counts the number of seconds since boot or roll-over.
 * + t_sample is a seconds count-down timer used by the main sampling loop to 
 *   time baro/temp sampling when not sleeping. Set by the foreground process
 *   and timed by the system timer (timer6).
 * + F_setup is a bit-mapped byte that controls the flow of setup with bits
 *   indicating what value setup works on: refer to the constants S_MK_SETUP
 *   in ms5607.h.
 * + The bit objects are flags that control program flow & user interaction.
 *      + "b_" flags indicate a button press - st (short), lg (long), rp 
 *        (repeat). They are set by the IOC section of the interrupt handler.
 *        All button flags except repeat must be cleared by the foreground 
 *        when processed. The repeat flag is cleared automatically by the 
 *        interrupt handler.
 *      + "f_" flags direct behavior:
 *          + f_sp_bg_ok indicates to the foreground when it is safe to sleep -
 *            e.g. there is no pending button press or timing activity.
 *          + f_dsleep is TRUE when the unit has awaken from deep sleep or boot.
 *            This is used to skip possible erroneous vertical speed & climb/
 *            descent accumulator calculations that might result in first sample
 *            from the baro sensor.
 *      + "d_" flags indicate what value is to be displayed.
 *      + u_metric is TRUE if the displayed unit is metric or (FALSE) english.
 *        It also determines what unit is used for setup.
 *      + "s_" flags are system fault flags that halt operation and force a 
 *        reset.
 * + baroCal holds the calibration constants & CRC data from the MS5607.
 * + baroDat holds the pressure & temperature read from the MS5607, the 
 *   compensation values, and the calculated results.
 * 
 *   *For both baroCal & baroDat all of the elements involved in calculations
 *   are set to long long to allow full resolution of the values and eliminate
 *   casting in the various formulas.
 * + The "msg_" objects are message strings displayed on the LCD.
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
uint8_t t_msc,
        t_sc,
        t_sample,
        F_setup;
__bit   b_st_up,
        b_rp_up,
        b_st_dn,
        b_rp_dn,
        b_lg_dn,
        f_sp_bg_ok,
        f_dsleep,
        d_alt,
        d_climb,
        d_descent,
        d_vspeed,
        d_temp,
        d_pres,
        d_baro,
        u_metric,
        s_mcu;

union {
    uint8_t b[sizeof(signed long long)];
    signed long long c;
} baroCal[S_SZ_MS5607_CAL];
struct {
    union {
        uint8_t             b[sizeof(signed long long)];
        signed long long    v;
    } D1;                                       //Sensor pressure sample
    union {
        uint8_t             b[sizeof(signed long long)];
        signed long long    v;
    } D2;                                       //Sensor temperature sample
    signed long long    dT;
    signed long long    T2;
    signed long long    OFF;                    //First order offset correction factor
    signed long long    OFF2;                   //Second order offset correction factor
    signed long long    SENS;                   //First order sensitivity correction factor
    signed long long    SENS2;                  //Second order sensitivity correction factor
    signed long long    P;                      //Pressure millibars
    signed long long    TEMP;                   //Temp C
    signed long         TEMP_F;                 //Temp F
    signed long         P_Hg;                   //Pressure inches of Hg
    signed long         bc_mbar;                //Baro correction millibars
    signed long         bc_Hg;                  //Baro correction inches of Hg
    signed long         alt_m;                  //Altitude meters
    signed long         alt_f;                  //Altitude feed
    signed long         alt_m_p;                //Previous sample altitude meters (accumulators)
    signed long         alt_f_p;                //Previous sample altitude feet  (accumulators
    signed long         alt_m_vsp;              //Previous sample altitude meters (vspeed calc)
    signed long         alt_f_vsp;              //Previous sample altitude feet (vspeed calc)
    signed long         climb_m;                //Accumulated climb, meters
    signed long         climb_f;                //Accumulated climb, feet
    signed long         descent_m;              //Accumulated descent, meters
    signed long         descent_f;              //Accumulated descent, feet
    signed long         vspeed_m;               //Rate of climb, M/minute
    signed long         vspeed_f;               //Rate of climb, F/minute
} baroDat;

unsigned char *msg_mcu=(unsigned char *)"MCU pr ^",
              *msg_crc=(unsigned char *)"CRC pr ^",
              *msg_sel_a=(unsigned char *)"s  Alt v",
              *msg_sel_b=(unsigned char *)"s Baro v",
              *msg_sel_t=(unsigned char *)"s Toff v";


/*Function main.  Firmware entry point from C root and foreground execution
 * control.
 * 
 * Initialization is done in two steps. First MCU peripherals are configured by
 * calling the initialize function. Then external peripherals are configured:
 *  + LCD first so that errors can be displayed
 *  + If the s_mcu status flag is set (by Initialize) an error message is 
 *    displayed and execution halted. The user can press the UP button to reset
 *    the MCU to attempt to clear the error.
 *  + The baro sensor calibration data is loaded and its validity checked via 
 *    the CRC4 value. A bad CRC will halt execution and display an error. The 
 *    user can press the UP button to reset the MCU to attempt to clear the 
 *    error.
 *  + The version & build number are displayed at the end of initialization.
 *    The user must press the UP button to start normal operation.
 * 
 * The RUN loop is an infinite loop and is the normal operating state and 
 * handles the display of barometric data and user interaction. A detailed
 * explanation of its operation can be found at the beginning of the various 
 * operating features.
 *
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void main(void)
{
    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * FUNCTION DATA.
     * + counter is a general purpose loop counter and scratch variable.
     * + baro_cmd holds a MS5607 command byte.
     * + dsp_incr is the value the setting display increments when adjusting
     *   baro compensation.
     * + t_off stores the number of contiguous iterations thru the RUN loop 
     *   without button or background activity to control deep sleep.
     * + baro_h & lcd_h are the spi handles used by the SPI module to 
     *   communicate with the MS5607 baro sensor & LCD module. Refer to spi.h.
     * + The "*dsp_" objects are used to access the appropriate value for
     *   display based on the u_metric object.
     * + The "dsp_bc_" objects hold the scaled values of their respective 
     *   elements in baroDat based on the u_metric object. These are used for
     *   display.
     * + The "lim_setup_" objects define the setting limits for baro
     *   compensation.
     * + The "baro_*xb" & "lcd_*xb" objects are the SPI receive & transmit 
     *   buffers for the MS5607 & LCD respectively.
     * + lcd_inib is a dedicated SPI transmit buffer initialized with the LCD
     *   configuration data.
     * + The "dsp_unit_" objects hold the unit display character for the LCD
     *   display value.
     * + t_dsleep is the number of contiguous sleep increments for the unit to
     *   enter deep sleep.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    uint8_t       counter=0,
                  baro_cmd,
                  dsp_incr,
                  hys_climb=0,
                  hys_descent=0;
    unsigned int  t_off=0;
    spi_h         baro_h;
    spi_h         lcd_h;
    signed long  *dsp_P,
                 *dsp_bc,
                 *dsp_alt,
                 *dsp_climb,
                 *dsp_descent,
                 *dsp_vspeed,
                 *dsp_TEMP,
                 *dsp_setup,
                  dsp_bc_Hg,
                  dsp_bc_mbar,
                  lim_setup_up,
                  lim_setup_dn,
                  t_dsleep=S_IN_OFF;
    unsigned char baro_rxb[S_SZ_MS5607_BUF];
    unsigned char baro_txb[S_SZ_MS5607_BUF];
    unsigned char lcd_rxb[S_SZ_LCD_BUF];
    unsigned char lcd_txb[S_SZ_LCD_BUF];
    unsigned char lcd_inib[S_SZ_LCD_BUF]={S_IN_LCD_FCTST1,S_IN_LCD_BIASET,\
        S_IN_LCD_PWRCTL,S_IN_LCD_FLWCTL,S_IN_LCD_CONSET,S_IN_LCD_FCTST0,\
        S_IN_LCD_DISPLY,S_IN_LCD_CLRDSP,S_IN_LCD_ENTRYM};
    char dsp_unit_P,
         dsp_unit_bc,
         dsp_unit_alt,
         dsp_unit_TEMP,
         dsp_unit_setup;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * PIC PERIPHERAL INITIALIZATION:
     * + Start charge pump & delay for stabilization
     * + Power-up Baro & LCD, delay for stabilization
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    Initialize();
    PIN_OD_CHGP=TRUE;
    __delay_ms(S_TM_INT_CHGP);
    PIN_OD_PW_LB=TRUE;
    __delay_ms(S_TM_INT_LCDB);


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * INITIALIZE LCD:
     * + Initialize SPI handle
     * + Call LCDINI macro to send configuration commands
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    lcd_h.rxb = lcd_rxb;
    lcd_h.csp = S_MK_CS_LCD;
    lcd_h.port = &PORTC;
    LCDINI();


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * TEST FOR MCU FAULT: fault detection is done in the Initialize function.
     * + Display failure message
     * + Halt & wait for key press
     * + Clear fault bits
     * + Reboot
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if (s_mcu)
    {
        lcd_h.txb=msg_mcu;
        lcd_h.szb=(unsigned char)strlen((const char *)msg_mcu);
        spi_m_TxData(&lcd_h);
        while (!(b_st_up));
        STKOVF=FALSE;
        STKUNF=FALSE;
        PCON0bits.nRWDT=TRUE;
        RESET();
    }


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * BARO DATA OBJECT INITIALIZATION:
     * + Initialize SPI handle
     * + Zero data objects - important for unused high bytes of long long 
     *   objects storing smaller long/int values
     * + Set standard atmosphere baro correction
     * + Reset MS5607 to ensure calibration data loaded from PROM
     * + Read MS5607 calibration data
     * + Check CRC and if invalid display error message & reboot
     *      + After the user presses the UP button the LCD & baro sensor are
     *        powered down with a delay before reboot to provide a 'hard' reset
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    baro_h.rxb = baro_rxb;
    baro_h.txb = baro_txb;
    baro_h.csp = S_MK_CS_BARO;
    baro_h.port = &PORTC;
    memset((void *) baroCal, (int) 0x00, (size_t) sizeof(baroCal));
    memset((void *) &baroDat, (int) 0x00, (size_t) sizeof(baroDat));
    baroDat.bc_Hg = (signed long) S_QY_ALT_HG_STD;
    baroDat.bc_mbar = (signed long) S_QY_ALT_MBAR_STD;

    //Start by resetting baro sensor
    BAROINI();

    //Read PROM coefficients: each is accessed by incrementing initial baro_cmd
    //value by two
    for (counter = 0, baro_cmd = S_CD_MS5607_C0; \
            (counter < S_SZ_MS5607_CAL); ++counter, baro_cmd += 2)
    {
        __delay_ms(S_TM_MS5607_CMD);
        memset((void *) baro_txb, (int) 0x00, S_SZ_MS5607_BUF); //Clear
        baro_txb[S_AR_MS5607_CMD] = baro_cmd;                   //Next value
        baro_h.szb = S_SZ_MS5607_CEF;
        spi_m_TxData(&baro_h);                                  //Retrieve
        //Store
        baroCal[counter].b[S_AR_UINT16_LL_L] = baro_rxb[S_AR_MS5607_CEF_L];
        baroCal[counter].b[S_AR_UINT16_LL_H] = baro_rxb[S_AR_MS5607_CEF_H];
    }

    //Check data validity with CRC
    if (crc4((signed long long *) baroCal))
    {
        lcd_h.txb=msg_crc;
        lcd_h.szb=(unsigned char)strlen((const char *)msg_crc);
        spi_m_TxData(&lcd_h);
        while (!(b_st_up));
        CTRLOFF();                              //Control pins off
        PIN_OD_PW_LB=FALSE;                     //Power down Baro & LCD
        __delay_ms(S_TM_CRC_RB);                //Delay for hard reset
        RESET();
    }


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * DISPLAY VERSION NUMBER:
     * + Format as major*10000+minor*100+build
     * + Display and wait for up button press start normal operation 
     * + Clear button flags to prevent display units/value/setup from engaging
     *   upon entering RUN loop (could occur if user presses down button before
     *   up)
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    fmtLCDdsp(S_MK_FMT_NON,'V','^',(char *)lcd_txb,\
        (signed long)((VERMAJOR*10000)+(VERMINOR*100)+VERBUILD));
    lcd_h.szb = S_SZ_LCD_CHAR;
    spi_m_TxData(&lcd_h);
    while (!(b_st_up));                         //Wait for up (^) button press
    b_st_up=FALSE;
    b_st_dn=FALSE;
    b_lg_dn=FALSE;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * RUN LOOP
     *  + General flow:
     *      + Display selection - what to display & in which unit (E/M)
     *      + Data acquisition - get new pressure/temperature
     *      + Apply compensation equations
     *      + Calculate ancillary data - altitude, etc.
     *      + Display data
     *      + Check for / perform setup (set baro compensation)
     *      + Sleep / deep sleep control
     *  + Watchdog is primarily used to wake-up from sleep for sample & display
     *    updates.
     +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    SWDTEN=TRUE;
    while (TRUE)
    {
        //Clear to prevent resets when not sleeping
        CLRWDT();

        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * DISPLAY SELECTION:
         * + Check for short up button press to select next value.
         * + Check for change in unit display value
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (b_st_up)
        {
            if (d_alt)
            {
                d_alt=FALSE;
                d_climb=TRUE;
            }
            else if (d_climb)
            {
                d_climb=FALSE;
                d_descent=TRUE;
            }
            else if (d_descent)
            {
                d_descent=FALSE;
                d_vspeed=TRUE;
            }
            else if (d_vspeed)
            {
                d_vspeed=FALSE;
                d_temp=TRUE;
            }
            else if (d_temp)
            {
                d_temp=FALSE;
                d_pres=TRUE;
            }
            else if (d_pres)
            {
                d_pres=FALSE;
                d_baro=TRUE;
            }
            else
            {
                d_alt=TRUE;
                d_baro=FALSE;
            }
            b_st_up=FALSE;
            t_sample=0;                         //Force update
        }
        if (b_st_dn)
        {
            b_st_dn=FALSE;
            u_metric^=TRUE;
            t_sample=0;                         //Force update
        }


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * DATA ACQUISITION AND DISPLAY.
         *  + This block includes acquisition, calculation, and display.
         *  + Updates occur when t_sample is clear: this will always be the case
         *    when waking from either sleep mode
         *  + t_sample is reset at the end of the display update to emulate the
         *    display update timing of sleep: this happens when the display back
         *    light is on or the user is pressing buttons
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (!(t_sample))
        {
            //A single clear for block to save space...
            memset((void *) baro_txb, (int) 0x00, S_SZ_MS5607_BUF);
            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * READ TEMPERATURE
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            //Acquisition
            baro_txb[S_AR_MS5607_CMD]=S_CD_MS5607_D2;
            baro_h.szb=S_SZ_MS5607_CNV;
            spi_m_TxData(&baro_h);
            __delay_ms(S_TM_MS5607_CMD);
            //Read
            baro_txb[S_AR_MS5607_CMD]=S_CD_MS5607_ADC;
            baro_h.szb=S_SZ_MS5607_ADC;
            spi_m_TxData(&baro_h);
            //Store
            baroDat.D2.b[S_AR_LONG_LL_1]=baro_rxb[S_AR_MS5607_CEF_1];
            baroDat.D2.b[S_AR_LONG_LL_2]=baro_rxb[S_AR_MS5607_CEF_2];
            baroDat.D2.b[S_AR_LONG_LL_3]=baro_rxb[S_AR_MS5607_CEF_3];

            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * READ PRESURE
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            //Acquisition
            baro_txb[S_AR_MS5607_CMD]=S_CD_MS5607_D1;
            baro_h.szb=S_SZ_MS5607_CNV;
            spi_m_TxData(&baro_h);
            __delay_ms(S_TM_MS5607_CMD);
            //Read
            baro_txb[S_AR_MS5607_CMD]=S_CD_MS5607_ADC;
            baro_h.szb=S_SZ_MS5607_ADC;
            spi_m_TxData(&baro_h);
            //Store
            baroDat.D1.b[S_AR_LONG_LL_1]=baro_rxb[S_AR_MS5607_CEF_1];
            baroDat.D1.b[S_AR_LONG_LL_2]=baro_rxb[S_AR_MS5607_CEF_2];
            baroDat.D1.b[S_AR_LONG_LL_3]=baro_rxb[S_AR_MS5607_CEF_3];

            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * FIRST ORDER CALCULATIONS: always done
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            baroDat.dT=(baroDat.D2.v-(baroCal[S_AR_MS5607_C5].c*S_RT_MS5607_C5));
            baroDat.TEMP=(2000+(baroDat.dT*baroCal[S_AR_MS5607_C6].c)/\
                    S_RT_MS5607_C6);
            baroDat.OFF=(baroCal[S_AR_MS5607_C2].c\
                    *S_RT_MS5607_C2)+((baroCal[S_AR_MS5607_C4].c*\
                    baroDat.dT)/S_RT_MS5607_C4dT);
            baroDat.SENS=(baroCal[S_AR_MS5607_C1].c*S_RT_MS5607_C1)+\
                    ((baroCal[S_AR_MS5607_C3].c*baroDat.dT)/S_RT_MS5607_C3dT);

            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * SECOND ORDER CALCULATIONS: for temps < 20/-15C
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            if (baroDat.TEMP < S_LM_MS5607_2ORD)
            {
                baroDat.T2=(((baroDat.dT * baroDat.dT))/S_RT_MS5607_dT);
                baroDat.TEMP-=baroDat.T2;
                baroDat.OFF2=(61*((baroDat.TEMP-S_LM_MS5607_2ORD)*\
                        (baroDat.TEMP-S_LM_MS5607_2ORD)))/S_RT_MS5607_OFF2;
                baroDat.SENS2=(2*((baroDat.TEMP-S_LM_MS5607_2ORD)*\
                        (baroDat.TEMP-S_LM_MS5607_2ORD)));

                //Compensate for very low temps (<-15C)
                if (baroDat.TEMP<(S_LM_MS5607_VLTMP*-1))
                {
                    baroDat.OFF2=baroDat.OFF2+(15*((baroDat.TEMP+\
                            S_LM_MS5607_VLTMP)*\
                            (baroDat.TEMP+S_LM_MS5607_VLTMP)));
                    baroDat.SENS2=baroDat.SENS2+(8*((baroDat.TEMP+\
                            S_LM_MS5607_VLTMP)*\
                            (baroDat.TEMP+S_LM_MS5607_VLTMP)));
                }
            }
            else
            {
                baroDat.OFF2=0;
                baroDat.SENS2=0;
            }

            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * COMPLETE TEMPERATURE COMPENSATED PRESSURE CALCULATION:
             * + Pressure mbar
             * + Pressure Hg
             * + Altitude feet
             * + Altitude meters
             * + Temperature fahrenheit 
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            baroDat.OFF-=baroDat.OFF2;
            baroDat.SENS-=baroDat.SENS2;
            baroDat.P=(((baroDat.D1.v*baroDat.SENS)/S_RT_MS5607_SENS)-\
                    baroDat.OFF)/S_RT_MS5607_P;
            baroDat.P_Hg=(signed long)(baroDat.P*S_RT_DECIMAL_3)/\
                    S_QY_ALT_MBAR_INHG;
            baroDat.alt_f=S_QY_ALT_HG_STD-((baroDat.P_Hg*S_RT_DECIMAL_1)-\
                    (baroDat.bc_Hg-S_QY_ALT_HG_STD));
            baroDat.alt_m=(signed long) (S_QY_ALT_MBAR_STD-\
                    ((baroDat.P*S_RT_DECIMAL_1)-(baroDat.bc_mbar-\
                    S_QY_ALT_MBAR_STD)))/S_QY_ALT_MBAR_METR;
            baroDat.TEMP_F=(signed long)((baroDat.TEMP*900)/500)+3200;

            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * CALCULATE VERTICAL SPEED & CLIMB/DESCENT ACCUMULATORS:
             * + The f_dsleep flag is set at boot and when waking from deep 
             *   sleep to prevent false measurements based on first sample
             *      + If the climb/descent objects are zero it is assumed that
             *        the unit just booted so the history objects are set to the
             *        current altitude to prevent a false accumulation
             *      + If they are not zero a change in altitude is considered 
             *        valid - the unit could have changed elevation while 
             *        sleeping
             * + Otherwise the vertical speed & climb descent accumulators are 
             *   calculated
             *      + Separate blocks are used to evaluate the minimum required
             *        change in altitude for flexibility based on operational
             *        experience & perceived accuracy
             * + The vertical speed history objects (..._vsp) are always updated
             *   since vertical speed is an instantaneous measurement
             * + The accumulators employ two levels of filtering to minimize 
             *   erroneous values due to spurious changes in altitude when
             *   the unit is idle. The first requires a minimum altitude change
             *   set by the constant S_LM_AC_S_MIN. The second is a simple
             *   hysteresis algorithm that requires a number of contiguous 
             *   samples (set by the constant S_LM_AC_HYS). A disruption in a
             *   sample set clears the hysteresis with no accumulator update and
             *   sampling starts over. The flow is a bit clunky and requires
             *   updates to the hysteresis objects (hys_climb/hys_descent) in
             *   opposing evaluation blocks. The macro RESETAHYS() is used to
             *   help ensure consistency along with a single if block to update
             *   the history (alt_x_p) objects.
             * + The accumulator history objects (..._p) are only updated when
             *   the accumulators are since they represent a cumulative change -
             *   a change in altitude less than criteria will not be lost if
             *   the trend continues in successive samples
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            if (f_dsleep)
            {
                f_dsleep=FALSE;                 //Skip on first sample
                RESETAHYS();                    //Reset hysteresis
                if (!(baroDat.climb_m | baroDat.descent_m))
                {
                    baroDat.alt_m_p=baroDat.alt_m;
                    baroDat.alt_f_p=baroDat.alt_f;
                }
            }
            else
            {
                //Calculate vertical speed if > than minimum per sample
                if (labs(baroDat.alt_m-baroDat.alt_m_vsp)>S_LM_VS_S_MIN)
                {
                    baroDat.vspeed_m=(baroDat.alt_m-baroDat.alt_m_vsp)*\
                            S_RT_VS_RATE;
                    baroDat.vspeed_f=(baroDat.alt_f-baroDat.alt_f_vsp)*\
                            S_RT_VS_RATE;
                }
                else                            //Otherwise ensure zero
                {
                    baroDat.vspeed_m=0;
                    baroDat.vspeed_f=0;
                }
                //Climb/descent accumulator calculation
                if (labs(baroDat.alt_m-baroDat.alt_m_p)>S_LM_AC_S_MIN)
                {
                    if (baroDat.alt_m>baroDat.alt_m_p)  //Climb
                    {
                        hys_descent=0;                  //Clear opposite
                        if (++hys_climb==S_LM_AC_HYS)
                        {
                            baroDat.climb_m+=(baroDat.alt_m-baroDat.alt_m_p);
                            baroDat.climb_f+=(baroDat.alt_f-baroDat.alt_f_p);
                        }
                    }
                    else                                //Descent
                    {
                        hys_climb=0;                    //Clear opposite
                        if (++hys_descent==S_LM_AC_HYS)
                        {
                            baroDat.descent_m+=(baroDat.alt_m_p-baroDat.alt_m);
                            baroDat.descent_f+=(baroDat.alt_f_p-baroDat.alt_f);
                        }
                    }
                    //Update previous altitudes for next & clear hysteresis
                    if ((hys_climb+hys_descent)==S_LM_AC_HYS)
                    {
                        RESETAHYS();
                        baroDat.alt_m_p=baroDat.alt_m;
                        baroDat.alt_f_p=baroDat.alt_f;
                    }
                }
                else                                    //If equal start over
                {
                    RESETAHYS();
                }
            }
            //Always update previous vertical speed for next sample test
            baroDat.alt_m_vsp=baroDat.alt_m;
            baroDat.alt_f_vsp=baroDat.alt_f;

            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * FORMAT DISPLAY STRING:
             * + First move cursor to start of display since this makes use of 
             *   the same buffer. The time required to format the output also 
             *   gives the LCD processor time to update.
             * + Display units assigned based on active system of measurement
             * + Display pointers assigned to the respective source objects
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            LCDHOME();
            if (u_metric)                       //Display in metric units
            {
                dsp_unit_P='m';
                dsp_unit_bc=dsp_unit_P;
                dsp_unit_alt='M';
                dsp_unit_TEMP='C';
                dsp_P=(signed long *) &baroDat.P;
                //Hack to eliminate 1000's decimal place
                dsp_bc_mbar=(baroDat.bc_mbar/S_RT_DECIMAL_1);
                dsp_bc=&dsp_bc_mbar;
                dsp_alt=&baroDat.alt_m;
                dsp_climb=&baroDat.climb_m;
                dsp_descent=&baroDat.descent_m;
                dsp_vspeed=&baroDat.vspeed_m;
                dsp_TEMP=(signed long *) &baroDat.TEMP;
            }
            else                                //Display in english units
            {
                dsp_unit_P='H';
                dsp_unit_bc=dsp_unit_P;
                dsp_unit_alt='F';
                dsp_unit_TEMP='F';
                dsp_P=(signed long *) &baroDat.P_Hg;
                //Hack to eliminate 1000's decimal place
                dsp_bc_Hg=(baroDat.bc_Hg/S_RT_DECIMAL_1);
                dsp_bc=&dsp_bc_Hg;
                dsp_alt=&baroDat.alt_f;
                dsp_climb=&baroDat.climb_f;
                dsp_descent=&baroDat.descent_f;
                dsp_vspeed=&baroDat.vspeed_f;
                dsp_TEMP=(signed long *) &baroDat.TEMP_F;
            }

            //Write display string to LCD SPI Tx buffer
            if (d_alt) fmtLCDdsp(S_MK_FMT_COM,'A',dsp_unit_alt,\
                    (char *)lcd_txb,*dsp_alt);
            else if (d_climb) fmtLCDdsp(S_MK_FMT_COM,'C',dsp_unit_alt,\
                    (char *)lcd_txb,*dsp_climb);
            else if (d_descent) fmtLCDdsp(S_MK_FMT_COM,'D',dsp_unit_alt,\
                    (char *)lcd_txb,*dsp_descent);
            else if (d_vspeed) fmtLCDdsp(S_MK_FMT_COM,'V',dsp_unit_alt,\
                    (char *)lcd_txb,*dsp_vspeed);
            else if (d_temp) fmtLCDdsp(S_MK_FMT_DEC,'T',dsp_unit_TEMP,\
                    (char *)lcd_txb,*dsp_TEMP);
            else if (d_pres) fmtLCDdsp(S_MK_FMT_NON,'P',dsp_unit_P,\
                    (char *)lcd_txb,*dsp_P);
            else if (d_baro) fmtLCDdsp(S_MK_FMT_NON,'B',dsp_unit_bc,\
                    (char *)lcd_txb,*dsp_bc);

            //Send to display
            lcd_h.szb = S_SZ_LCD_CHAR;
            spi_m_TxData(&lcd_h);

            //Reset timer for next display update if not sleeping
            t_sample=S_TM_SAMPLE;
        }


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * SETUP:
         * + This block is entered when a long down button press is detected
         * + If the climb or descent accumulator is displayed the accumulators
         *   are cleared and setup is exited: by not clearing b_lg_dn the block
         *   exits
         * + The user is first prompted to select barometric compensation in 
         *   altitude or corrected sea level pressure, or the deep sleep 
         *   timeout (up button selects)
         * + The measurement unit (E/M) used for baro or altitude is the one 
         *   active prior to entering setup
         * + The down button confirms the choice and proceeds to setting the
         *   value
         * + A long down button press exits setup without changes
         * + Watchdog (WDT) is kept active throughout setup so must be cleared
         *   when sitting in selection loops to prevent reset
         * + The bit-mapped byte object F_setup controls what value setup
         *   operates on
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if (b_lg_dn)
        {
            if (d_climb | d_descent)            //Clear accumulators
            {
                baroDat.climb_m=0;
                baroDat.climb_f=0;
                baroDat.descent_m=0;
                baroDat.descent_f=0;
                baroDat.alt_m_p=baroDat.alt_m;  //Prevent false accumulation
                baroDat.alt_f_p=baroDat.alt_f;
            }
            else
            {
                F_setup=S_MK_SETUP_FIRST;       //Set to first setup item
                b_st_up=FALSE;                  //Clear all button press flags
                b_st_dn=b_st_up;
                b_lg_dn=b_st_up;
            }

            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * Select value to set
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            while (!(b_st_dn | b_lg_dn))
            {
                CLRWDT();
                LCDHOME();
                lcd_h.szb = S_SZ_LCD_CHAR;
                if (F_setup & S_MK_SETUP_ALT) strcpy((char *)lcd_txb,\
                        (const char *)msg_sel_a);
                else if (F_setup & S_MK_SETUP_BAR)
                    strcpy((char *)lcd_txb,(const char *)msg_sel_b);
                else if (F_setup & S_MK_SETUP_TOF)
                    strcpy((char *)lcd_txb,(const char *)msg_sel_t);
                spi_m_TxData(&lcd_h);
                while (!(b_st_up | b_st_dn | b_lg_dn)) CLRWDT();
                if (b_st_up)
                {
                    if (F_setup & S_MK_SETUP_LAST) F_setup=S_MK_SETUP_FIRST;
                    else F_setup=F_setup>>1;
                    b_st_up=FALSE;
                }
            }

            /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
             * If setup wasn't aborted by user (long down button press), setup 
             * display and limits set value.
             * 
             *  + Object F_setup indicates setup of altitude, baro, or deep 
             *    sleep timeout.
             *  + Object u_metric specifies units.
             *  + Object dsp_unit_setup stores the postfix character to display
             *    on the LCD (e.g. M, F, etc.).
             *  + Object dsp_setup points to the display value.
             +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
            if (!(b_lg_dn))
            {
                b_st_dn=FALSE;
                if (F_setup & S_MK_SETUP_ALT)       //Use altitude (elevation)
                {
                    dsp_unit_setup=dsp_unit_alt;
                    dsp_setup=dsp_alt;
                    if (u_metric)                   //Assign metric limits
                    {
                        lim_setup_up=S_LM_ALT_M_UP;
                        lim_setup_dn=S_LM_ALT_M_DN;
                    }
                    else                            //Assign english limits
                    {
                        lim_setup_up=S_LM_ALT_F_UP;
                        lim_setup_dn=S_LM_ALT_F_DN;
                    }
                }
                else if (F_setup & S_MK_SETUP_BAR)  //Use baro compensation
                {
                    dsp_unit_setup=dsp_unit_bc;
                    dsp_setup=dsp_bc;
                    if (u_metric)                   //Assign metric limits
                    {
                        lim_setup_up=S_LM_BAR_M_UP;
                        lim_setup_dn=S_LM_BAR_M_DN;
                    }
                    else                            //Assign english limits
                    {
                        lim_setup_up=S_LM_BAR_H_UP;
                        lim_setup_dn=S_LM_BAR_H_DN;
                    }
                }
                else if (F_setup & S_MK_SETUP_TOF)  //Set time off
                {
                    dsp_unit_setup=S_CH_CAP_M;
                    t_dsleep=(t_dsleep*S_TM_SAMPLE)/S_QY_SECINMIN;
                    dsp_setup=&t_dsleep;
                    lim_setup_up=S_LM_TOF_UP;
                    lim_setup_dn=S_LM_TOF_DN;
                }

                /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                 * Display & let user set value.
                 *  + Up & down buttons (short press) increment/decrement.
                 *  + Holding a button down engages repeat to more quickly 
                 *    update a large setting deviation.
                 *  + A long down button press saves the new value and exits 
                 *    setup.
                 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
                for (dsp_incr=1; (!(b_lg_dn)); )
                {
                    CLRWDT();
                    LCDHOME();
                    lcd_h.szb = S_SZ_LCD_CHAR;
                    fmtLCDdsp(S_MK_FMT_NON,'s',dsp_unit_setup,(char *)lcd_txb,\
                            *dsp_setup);
                    spi_m_TxData(&lcd_h);       //Update display value

                    //Wait for button press
                    while (!(b_st_up | b_st_dn | b_lg_dn | b_rp_up | b_rp_dn))
                        CLRWDT();

                    if (b_rp_up | b_rp_dn)      //Set repeat value
                    {
                        dsp_incr=S_QY_INCR_RP;
                        __delay_ms(S_TM_INCR_RP);
                    }
                    else dsp_incr=S_QY_INCR_ST;

                    if (b_st_up | b_rp_up)      //Increment up
                    {
                        b_st_up=FALSE;
                        if ((*dsp_setup+dsp_incr)<=lim_setup_up) \
                            *dsp_setup+=dsp_incr;
                    }
                    else if (b_st_dn | b_rp_dn) //Increment down
                    {
                        b_st_dn=FALSE;
                        if ((*dsp_setup-dsp_incr)>=lim_setup_dn) \
                            *dsp_setup-=dsp_incr;
                    }
                }
                /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
                 * Complete setup - store values
                 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
                if (F_setup & S_MK_SETUP_ALT)           //Altitude
                {
                    if (u_metric)
                    {
                        baroDat.bc_mbar=S_QY_ALT_MBAR_STD-((S_QY_ALT_MBAR_STD-\
                                (*dsp_setup*S_QY_ALT_MBAR_METR))-\
                                ((signed long)baroDat.P*S_RT_DECIMAL_1));
                        baroDat.bc_Hg=(baroDat.bc_mbar*S_RT_DECIMAL_3)\
                                /S_QY_ALT_MBAR_INHG;
                    }
                    else
                    {
                        baroDat.bc_Hg=S_QY_ALT_HG_STD-((S_QY_ALT_HG_STD-*\
                                dsp_setup)-(baroDat.P_Hg*S_RT_DECIMAL_1));
                        baroDat.bc_mbar=(baroDat.bc_Hg*S_QY_ALT_MBAR_INHG)/\
                                S_RT_DECIMAL_3;
                    }
                }
                else if (F_setup & S_MK_SETUP_BAR)      //Barometric pressure
                {
                    if (u_metric)
                    {
                        baroDat.bc_mbar=(*dsp_setup*S_RT_DECIMAL_1);
                        baroDat.bc_Hg=(baroDat.bc_mbar*S_RT_DECIMAL_3)/\
                                S_QY_ALT_MBAR_INHG;
                    }
                    else
                    {
                        baroDat.bc_Hg=(*dsp_setup*S_RT_DECIMAL_1);
                        baroDat.bc_mbar=(baroDat.bc_Hg*S_QY_ALT_MBAR_INHG)/\
                                S_RT_DECIMAL_3;
                    }
                }
                else if (F_setup & S_MK_SETUP_TOF)      //Deep sleep timeout
                {
                    t_dsleep=(t_dsleep*S_QY_SECINMIN)/S_TM_SAMPLE;
                }
            }

            b_lg_dn=FALSE;                  //Ensure clear when exiting setup
            t_sample=0;                     //Clear to force update
            continue;                       //Skip sleep control for immediate update
        }


        /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
         * SLEEP CONTROL: controls sleep timing and determines when it is safe 
         * to sleep / deep sleep. In order for any sleep to occur there must not
         * be any background activity and no pending buttons to process.
         * + Timer t_off maintains a count of the number of sleep cycles: 
         *   button activity clears it
         * + When t_off equals t_dsleep, deep sleep is initiated; if t_dsleep is
         *   zero deep sleep is disabled
         * + The unit is placed into sleep between samples with baro & LCD 
         *   powered
         * + Wake from sleep is via WDT
         * + Wake from deep sleep is via IOC (e.g. user presses button)
         * + A new sample / display update is always done when waking from 
         *   either sleep by clearing t_sample
         +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
        if ((f_sp_bg_ok) && !(b_st_up | b_st_dn | b_lg_dn))
        {
            if ((++t_off==t_dsleep)&&(t_dsleep))    //Sleep or deep sleep
            {
                t_off=0;                    //Clear turn-off timer
                SPIDISABLE();               //Disable SPI to prevent current sink on SDO
                CTRLOFF();                  //Control pins off to prevent current sink
                PIN_OD_PW_LB=FALSE;         //Power down Baro & LCD
                PIN_OD_CHGP=FALSE;          //Charge pump off
                SWDTEN=FALSE;               //Prevent wake via watchdog
            }
        }
        else
        {
            t_off=0;                        //If not safe reset deep sleep timer
            continue;                       //Skip bottom of loop to avoid sleep
        }

        //Initiate either kind of sleep
        SLEEP();                            //Sleep entered when no KB or background activity
        SWDTEN=TRUE;                        //WDT always set at bottom of loop
        t_sample=0;                         //Update display waking from sleep

        //Wake from deep sleep by checking state of Baro/LCD power pin
        if (!(PIN_OD_PW_LB))
        {
            PIN_OD_CHGP=TRUE;               //Charge pump on
            __delay_ms(S_TM_INT_CHGP);      //Delay for stabilization
            PIN_OD_PW_LB=TRUE;              //Power-up LCD & baro sensor
            CTRLON();                       //Control pins returned to operating state
            SPIENABLE();                    //Enable SPI peripheral
            __delay_ms(S_TM_INT_LCDB);      //Wait for Baro sensor & LCD to boot
            BAROINI();                      //Ensure calibration data loads
            LCDINI();                       //Initialize LCD
            f_dsleep=TRUE;                  //Skip rate & accumulator calcs
        }
    }
} //End function main


/* Function fmtLCDdsp. Formats display data to a buffer in a standardized way
 * for display on the LCD. Data is formatted to "D 99999U" where:
 *      D=Prefix for the displayed value such as 'A' for altitude, 'T' for
 *        temperature, etc. 'D' is always displayed in the first (left) LCD
 *        position.
 *      9=numeric value to display, always right justified (padded with leading
 *        white space. The value can be optionally displayed with a thousands
 *        separator or decimal point (but not both). Values without either can
 *        display up to six digits; five with either one. Negative values are
 *        displayed with a leading '-' sign and limit the displayed value to
 *        five positions or less (if comma/decimal specified). Values exceeding
 *        the alloted display size will have trailing digits truncated.
 *      U=Postfix for the displayed value indicating the units (F for feet,
 *        mbar for millibars, etc.). 'U' is always displayed in the eighth 
 *        (right) LCD position.
 * 
 * Returns:     buf contains the formatted string. It is always eight bytes
 *              and is NOT NULL terminated.
 *
 * Parameters:  fmt=bit map indicating thousands separator/decimal point use:
 *                  S_MK_FMT_COM=thousands separator
 *                  S_MK_FMT_DEC=decimal point
 *                  S_MK_FMT_NON=none
 *              pre=character to display in first position
 *              post=character to display in last position
 *              buf=pointer to buffer to place formatted string in
 *              num=integer to format to display buffer buf
 *
*/
void fmtLCDdsp(unsigned char fmt, char pre, char post, char *buf, \
        signed long num)
{
    uint8_t      bp,                        //Offset in buf pointer
                 c,                         //Loop counter
                 pad;                       //# white space in front of number
    static __bit nf,                        //Negative flag
                 cf,                        //Comma flag
                 df;                        //Decimal flag
    static char  numascii[12];              //Buffer for ASCII representation of num

    //Set decimal point flag
    if (fmt&S_MK_FMT_DEC) df=TRUE; else df=FALSE;
    //If negative (two's compliment), flip to (+) and set nf flag
    if (num<0)
    {
        num^=0xFFFFFFFFFFFFFFFF;
        ++num;
        nf=TRUE;
    }
    else nf=FALSE;

    //Hack to handle a zero value with(out) decimal requirement
    memset((void *) numascii, (int) S_CH_ZERO, S_SZ_LCD_CHAR);

    //Write number as ASCII to internal buffer
    for (c=0; ((num) && (c < 10)); ++c)
    {
        numascii[c] = (num % 10) + '0';
        num /= (signed long) 10;
    }

    //If num=0 increment c to display a zero (decimal flag overrides next)
    if (!(c)) ++c;
    //Set decimal flag & adjust for minimum display of '0.00'
    if ((c<3)&&(df)) c=3;
    //Set comma flag
    if ((c>3)&&(fmt&S_MK_FMT_COM)) cf=TRUE; else cf=FALSE;
    //Add leading negative sign if set
    if (nf) numascii[c++]='-';
    //Adjust pad for display of comma or decimal
    if ((cf) | (df)) pad=c+1; else pad=c;
    //If pad>6 then out of room: otherwise calculate leading white space
    if (pad>6) pad=0;
    else pad=S_SZ_LCD_CHAR-(pad+2);

    //Add spaces after pre to right justify number against post
    for (bp=1,*buf=pre; (pad); --pad,++bp) *(buf+bp)=S_CH_SPACE;

    //Write number to buffer along with comma/decimal formatting
    for (; ((c)&&(bp<(S_SZ_LCD_CHAR-1))); --c,++bp)
    {
        if ((c==2)&&(df))
        {
            *(buf+bp)=S_CH_DECIMAL;
            ++c;                            //Adjust offset so digit isn't missed
            df=FALSE;
        }
        else if ((c==3)&&(cf))
        {
            *(buf+bp)=S_CH_COMMA;
            ++c;                            //Adjust offset so digit isn't missed
            cf=FALSE;
        }
        else *(buf+bp)=numascii[c-1];
    }

    //Write post character to eighth position of buffer
    *(buf+bp)=post;

    return;
} //End function fmtLCDdsp


/* Function crc4. Applies TE's CRC4 algorithm to the calibration data retrieved
 * from the MS5607 baro sensor and compares it to the 4-LSB value stored in the
 * CRC word. Adapted from TE application note AN520.
 * 
 * Although the PROM data is retrieved as 16-bit integers this function expects
 * them to be stored in 64-bit (signed long long) objects.
 * 
 * Returns:     0 if the checksums match
 *              non-zero if mismatch
 *
 * Parameters:  n_prom=pointer to array of eight unsigned integers (16-bit)
 *              retrieved from baro sensor's PROM BUT stored in signed long long
 *              objects. Integers must be ordered:
 *                  OEM, C1, C2, C3, C4, C5, C6, CRC
 *
*/
unsigned char crc4(signed long long *n_prom)
{
    unsigned char    cnt;                       //Current byte in 16-bit x 8 array
    unsigned int     n_rem;                     //CRC remainder
    signed long long crc_read;                  //Saves original crc value
    unsigned char    n_bit;                     //Counts bit position of current byte

    crc_read=*(n_prom+7);                       //save CRC
    *(n_prom+7)&=0xFF00;                        //Remove CRC to calculate

    for (n_rem=0, cnt=0; cnt<16; cnt++)
    {
        // choose LSB or MSB
        if (cnt%2) n_rem ^= (unsigned int) (*(n_prom+(cnt>>1)) & 0x00FF);
        else n_rem ^= (unsigned int) (*(n_prom+(cnt>>1))>>8);
        for (n_bit=8; n_bit>0; n_bit--)
        {
            if (n_rem & (0x8000)) n_rem=(n_rem << 1) ^ 0x3000;
            else n_rem=(n_rem << 1);
        }
    }

    n_rem=(n_rem >> 12);
    *(n_prom+7)=crc_read;
    return ((unsigned char)((crc_read&0x000F)^n_rem));
} //End function crc4


/* Function InterruptExec.  MCU interrupt handler.  Manages all enabled
 * interrupts which should be ordered according to priority high-low:
 *  + Timer6 (system timer)
 *  + IOC (Interrupt On Change)
 * 
 * Timer6 is configured as a non-precision system timer based on the MCU's 
 * system oscillator (Fosc) which in turn is derived from the HFINTOSC. The 
 * timer6 interrupt provides a 10mS time base which is used for various timing
 * activities.
 * 
 * A general timer is provided by the module objects t_msc & t_sc
 * which provide millisecond & seconds timing respectively. t_msc counts in 10mS
 * intervals to a maximum of 100 (1S) and is then reset. t_sc starts at boot
 * (epoch) and counts without to roll-over. These objects can be referenced by
 * code outside of the interrupt handler but should not be modified.
 * 
 * The timer6 interrupt is also used to provide timing for specific functions:
 *  + The button debounce timers, t_db_up & t_db_dn, are decremented in 10mS
 *    intervals to detect a valid button press and differentiate between long &
 *    short presses as well as support for the repeat function. Repeat is 
 *    defined as a button that is held down beyond the limit of the debounce
 *    timer.
 *  + The sample / display update timer t_sample is decremented in 1S increments.
 *    It is set by the RUN loop.
 *  + LCD back light timing & turn-off is managed by decrementing t_bl_on. It is
 *    set in the IOC section of this function.
 * 
 * Button press detection is based on the IOC feature of the PIC. Both positive
 * & negative edge IOC interrupts must be enabled. The inactive state of a 
 * button is electrical low (FALSE). Weak pull-down resistors (~10K) should be
 * used and, optionally in parallel 100nF MLCCs for added hysteresis. A button
 * press should pull the respective pin to electrical high (TRUE) by connecting
 * to Vdd thru a ~ 1K resistor.
 * 
 * The button press detection logic can differentiate between short & long
 * presses and a button that is held down (repeat). The detection logic relies
 * on timing support provided by the timer6 handler.
 * 
 * When a valid button press is detected the respective module-level button flag
 * (b_...) is set. Short (..._st_...) and long (..._lg_...) flags must be reset
 * by the foreground when processing is complete; repeat (..._rp_...) flags are
 * reset by the IOC handler when the button is released and should not be 
 * modified by foreground processes.
 * 
 * This function also sets the module-level status flag f_sp_bg_ok to indicate
 * when it is safe to sleep. This is defined as:
 *  + No pending debounce activity (e.g. buttons aren't depressed)
 *  + LCD back light is off
 * 
 * It is expected that the foreground (which controls sleep) will respect this
 * flag since timing services (via timer6) are not available when the PIC is in
 * sleep mode.
 * 
 * Continued button activity will reset the LCD back light timer if it is on.
 * This is done on the assumption that the user needs to see the display.
 *
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void __interrupt() InterruptExec(void)
{
    static __bit    f_db_up,                //TRUE when debounce detection active
                    f_db_dn;                //TRUE when debounce detection active
    static uint8_t  t_db_up=0,              //Debounce timer
                    t_db_dn=0,              //Debounce timer
                    t_bl_on=0;              //LCD on back light timer

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * SYSTEM TIMER. Non-precision timed background service. ~10 mS frequency.
     * Maintains timing objects:
     * + Decrement debounce timers if active
     * + Set repeat flags TRUE if respective debounce timer is expired & pin
     *   button TRUE
     * + t_msc counts 10mS increments
     * + t_sc counts elapsed seconds since wrap or when cleared
     * + Decrement sample timer if active
     * + Decrement LCD back light timer if active & turn off back light when 0
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if (TMR6IE & TMR6IF)
    {
        if (t_db_up) --t_db_up;
        else if (PIN_ID_UP & f_db_up) b_rp_up=TRUE;
        if (t_db_dn) --t_db_dn;
        else if (PIN_ID_DOWN & f_db_dn) b_rp_dn=TRUE;

        if (++t_msc==S_QY_TMR6_1S)
        {
            t_msc=0;
            ++t_sc;
            if (t_sample) --t_sample;
            if (t_bl_on) if (!(--t_bl_on)) PIN_OD_BL_LCD=FALSE;
        }
        TMR6IF=FALSE;
    }

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * IOC HANDLER: used for button press detection. Uses nested waterfall
     * if/else block.
     * 
     * When an IOC is detected the pin is first tested for a FALSE (low)
     * condition - which indicates the button was released. The debounce timer
     * is then tested to interpret the button press as invalid, short, long,
     * or expired. If the latter all control objects are cleared along with the
     * repeat button flag; otherwise the appropriate button press flag is set.
     * Note that the up long button press activates the LCD backlight & timer
     * and does not set a button press flag.
     * 
     * If the detected pin tests TRUE (high) the Baro/LCD power pin 
     * (PIN_OD_PW_LB) is first checked to ensure that it is TRUE (e.g. powered 
     * on): if not the button press is ignored which prevents the display from 
     * cycling to the next value when waking from deep sleep. Otherwise, the 
     * debounce timer & flag are set and a new button detection sequence starts.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if (IOCIE & PIN_IOC_UP)                     //UP button press
    {
        if (!(PIN_ID_UP))
        {
            if (t_db_up>S_TM_DB_MIN)            //Invalid (too short) - ignore
            {
                t_db_up=0;
                f_db_up=FALSE;
            }
            else if (t_db_up>S_TM_DB_ST)        //Short press
            {
                f_db_up=FALSE;
                t_db_up=0;
                b_st_up=TRUE;
            }
            else if (t_db_up>S_TM_DB_LG)        //Long press
            {
                f_db_up=FALSE;
                t_db_up=0;
                t_bl_on=S_TM_LCD_BKLT;
                PIN_OD_BL_LCD=TRUE;
            }
            else                                //Expired - reset
            {
                t_db_up=0;
                f_db_up=FALSE;
                b_rp_up=f_db_up;
            }
        }
        else if (PIN_OD_PW_LB)                  //If LCD on start new press detection
        {
            f_db_up=TRUE;
            t_db_up=S_TM_DB;
        }
        PIN_IOC_UP=FALSE;
    }
    if (IOCIE & PIN_IOC_DOWN)                   //DOWN button press
    {
        if (!(PIN_ID_DOWN))
        {
            if (t_db_dn>S_TM_DB_MIN)            //Invalid (too short) - ignore
            {
                t_db_dn=0;
                f_db_dn=FALSE;
            }
            else if (t_db_dn>S_TM_DB_ST)        //Short press
            {
                f_db_dn=FALSE;
                t_db_dn=0;
                b_st_dn=TRUE;
            }
            else if (t_db_dn>S_TM_DB_LG)        //Long press
            {
                f_db_dn=FALSE;
                t_db_dn=0;
                b_lg_dn=TRUE;
            }
            else                                //Expired - reset
            {
                t_db_dn=0;
                f_db_dn=FALSE;
                b_rp_dn=f_db_dn;
            }
        }
        else if (PIN_OD_PW_LB)                  //If LCD on start new press detection
        {
            f_db_dn=TRUE;
            t_db_dn=S_TM_DB;
        }
        PIN_IOC_DOWN=FALSE;
    }

    //Test if okay to sleep
    if (!(f_db_dn | f_db_up) && !(t_bl_on)) f_sp_bg_ok=TRUE;
    else f_sp_bg_ok=FALSE;

    //Reset backlight timer if new key activity while backlight on
    if ((f_db_dn | f_db_up) && (t_bl_on)) t_bl_on=S_TM_LCD_BKLT;

    return;
} //End function InterruptExec


/* Function Initialize. Should be called immediately after boot by function 
 * main. The bulk of MCU and PIC peripheral configuration should be done here: 
 * mostly for convention but technically not required. Sequence:
 *  + Configure system clock & watchdog timer
 *  + Boot detection and NVM load/validation
 *      + Test for reboot due to stack, WDT, etc.
 *      + Load data from NVM & validate (if used)
 *  + Initialize module control data
 *  + Configure I/O ports
 *  + Configure PPS pins
 *  + Configure PIC peripherals
 *  + Finish - start system timers & enable interrupts
 * 
 * Returns:  none
 *
 * Parameters:  none
 *
*/
void Initialize(void)
{
    di();                               //No interrupts during initialization

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * System clock & watchdog config
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    OSCCON1bits.NOSC = CFG_OSC_NOSC;
    OSCCON1bits.NDIV = CFG_OSC_NDIV;
    OSCFRQ = CFG_OSC_HFFRQ;
    while (!(OSCSTAT1bits.HFOR));       //Wait for oscillator to stabilize
    OSCTUNE = CFG_OSC_OSCTUNE;          //Shift oscillator frequency if necessary
    WDTCON = CFG_WDT_CON;               //Watchdog


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * BOOT & NVM PROTECTION AND MANAGEMENT.
     *
     * Recovering from CPU/firmware failures (e.g. watchdog, stack OF/UF, etc.)
     * and NVM corruption is critical for safety and field reliability.  While
     * neither should happen unless there's a true failure, statistically it's 
     * possible when operating in environmental extremes or the more likely
     * existence of a firmware bug.
     * 
     * The goal is to maintain safe operation, keep the unit intact, and record
     * the failure state (in that order).
     * 
     * BOOT PROCESS
     * The CPU boot status is always checked first to ensure that any abnormal
     * condition is identified before being cleared by subsequent execution.  
     * The PCON & STATUS registers are examined for reboot due to watchdog, 
     * stack OF/UF, and possibly other conditions.  If found the status flag
     * s_mcu is set and it's expected that function main halts normal operation.
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    if (((STKOVF) || (STKUNF)) || (!((PCON0bits.nRWDT) & (STATUSbits.nTO))))
        s_mcu=TRUE;
    else s_mcu=FALSE;
    

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * INITIALIZE MODULE CONTROL DATA
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    t_msc=0;
    t_sc=0;
    t_sample=0;
    b_st_up=FALSE;
    b_rp_up=FALSE;
    b_st_dn=FALSE;
    b_rp_dn=FALSE;
    b_lg_dn=FALSE;
    f_sp_bg_ok=FALSE;
    f_dsleep=TRUE;
    d_alt=TRUE;
    d_climb=FALSE;
    d_descent=FALSE;
    d_vspeed=FALSE;
    d_temp=FALSE;
    d_pres=FALSE;
    d_baro=FALSE;
    u_metric=FALSE;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * SET PORT CONFIGURATION:
     * + Set digital latch initial states
     * + Pin in/out-put states
     * + Analog / digital
     * + Weak pull-ups
     * + Open drain (tri-state)
     * + Slew rate
     * + Logic input levels
     * + Interrupt On Change (IOC)
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PORTA=S_IN_PORTA;
    PORTC=S_IN_PORTC;
    //Port A
    TRISA=IO_PORTA;
    ANSELA=AD_PORTA;
    WPUA=WP_PORTA;
    ODCONA=OD_PORTA;
    SLRCONA=SR_PORTA;
    INLVLA=IL_PORTA;
    IOCAP=IOC_PORTA_P;
    IOCAN=IOC_PORTA_N;
    //Port C
    TRISC=IO_PORTC;
    ANSELC=AD_PORTC;
    WPUC=WP_PORTC;
    ODCONC=OD_PORTC;
    SLRCONC=SR_PORTC;
    INLVLC=IL_PORTC;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * MAP PERIPHERAL PINS:
     * + SPI output pins
     * + SPI input pins
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PPS_P_SPI_CLK=PPS_M_SPI_CLK;
    PPS_P_SPI_SDO=PPS_M_SPI_SDO;
    SSP1DATPPS=PPS_P_SPI_SDI;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * CONFIGURE MSSP as master SPI
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    SPI_M_CONFIG();


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * CONFIGURE TIMERS:
     * 6=ISR timing service
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PR6=CFG_TMR_6_PR;
    T6CON=CFG_TMR_6_CON;
    

    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * PERIPHERAL MODULE DISABLE (energy conservation)
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PMD0=CFG_PMD_0;
    PMD1=CFG_PMD_1;
    PMD2=CFG_PMD_2;
    PMD3=CFG_PMD_3;
    PMD4=CFG_PMD_4;
    PMD5=CFG_PMD_5;


    /*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
     * INITIALIZATION COMPLETE:
     * + Enable peripheral & global interrupts
     * + Enable peripheral interrupts serviced by ISR
     * + Start system timer
     *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
    PEIE=ENABLE;                        //Enable peripheral interrupts
    ei();                               //Enable interrupts
    TMR6IE=TRUE;
    TMR6ON=TRUE;                        //System timer
    IOCIE=TRUE;

    return;
} //End function Initialize



//End of file