/*

DESCRIPTION:    PIC SPI service routines.

KNOWN BUGS:     functions are not reentrant. Attempting to address multiple 
                devices on the I2C bus at the same time (e.g. mixed calls for 
                start, tx, rx, etc.) is not supported and will have 
                indeterminate results. In other words, only communication with a
                 single device at a time is supported.

COPYRIGHT:      (c) 2020, Brian Cornell, all rights reserved.

DEPENDENCIES:

VERSION:        1.0

MODIFICATION HISTORY:

    9 Dec 20, BWC, first release
    

*/


#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <pic.h>
#include <xc.h>
#include <binary.h>
#include "spi.h"


/*++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 *                              Module Data
 * + spi_errno=error code set by last function call
 *++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
uint8_t spi_errno = SPI_ERR_NONE,
        count;


/* Function spi_m_TxData. * MASTER MODE *
 * 
 * Transmits & receives data to/from a slave. The same number of bytes are 
 * received as those transmitted.
 * 
 * Returns:     n/a
 *              
 *
 * Parameters:  handle=pointer to an spi handle object. Refer to spi.h for how
 *              to initialize.
 *
*/
void spi_m_TxData(spi_h *handle)
{
    //Make sure ready for Tx - ensure buffer & completion IF clear
    //Write to first byte of Rx buffer - will over-write with 1st valid byte
    *handle->rxb = SSP1BUF;
    SSP1IF = SPI_CLR;

    //Select device: set CS low
    *handle->port &= (handle->csp ^ S_MK_CS_ON);
    
    for (count = 0; (count < handle->szb); ++count)
    {
        SSP1BUF = *(handle->txb + count);           //Load Tx byte
        while (!(SSP1IF));                          //Wait for completion
        *(handle->rxb + count) = SSP1BUF;           //Load Rx byte
        SSP1IF = SPI_CLR;                           //Clear completion IF
    }

    //Deselect device: set CS high
    *handle->port |= handle->csp;

    return;
} //End function spi_m_TxData



//End of file